/*****************************************************************************
* Copyright (C) 2006, Mathias Kende                                          *
* licence@sectionpc.info                                                     *
*                                                                            *
* Le prsent logiciel est distribu sous la licence SectionPC dont vous avez *
* du avoir une copie.  Si ce n'est pas le cas,  vous pouvez vous en procurer *
* une  l'adresse http://licence.sectionpc.info  ou par courriel  l'adresse *
* ci-dessus.                                                                 *
*****************************************************************************/

#include "TPBD.h"
#include "TPBD_intern.h"

static char error[255];

/*
  Un bogue dans une version du programme fait que pour une base de N enregistrement, mme s'ils sont tous accessible,
  vous ne pouviez en enregistrer que N-1 au maximum.
  Le bogue a t corrig, mais pour des raisons de compatibilit avec une version prcdente, vous pouvez
  dcommenter la lignes ci-dessous pour ractiver le bogue.
*/
//#define LEGACY_MODE


//////////////////////////////////////////////////
//// FONCTIONS D'INTERFACE AVEC L'UTILISATEUR ////
//////////////////////////////////////////////////

int DBExist(char* DB)
{
  FILE *temp_file=NULL;
  if (fopen_s(&temp_file,DB,"r"))
    return FILE_DOES_NOT_EXIST;
  else
  {
    int res=(int)fseek(temp_file,0,SEEK_SET);
    if (res!=0)
      throw ERROR_SEEK_ERROR;
    int32_t val;
    res=(int)fread(&val,sizeof(int32_t),1,temp_file);
    if (res<1)
      throw ERROR_READ_ERROR;
    closef(temp_file);
    if (val==IS_DB)
      return FILE_IS_DB;
    else
      return FILE_IS_NOT_DB;
  }
}

void CreateDB(char* DB,int nbItems)
{
  DBase db;
  int exist=DBExist(DB);
  if (exist!=FILE_DOES_NOT_EXIST)
    throw ERROR_FILE_ALREADY_EXIST;
  if (fopen_s(&db.fic,DB,"wb+"))
    throw ERROR_COULD_NOT_CREATE_FILE;
  db.header.NbFAT=1;
  db.header.NbLT=1;
  db.header.NbMaxItem=(int32_t)nbItems;
  LT_Item lt_item;
  FAT_NbItems fat_nbitems;
  FAT_Item fat_item;
  lt_item.FatOffset=-1;
  fat_nbitems=0;
  fat_item.MemAlloc=0;
  fat_item.MemUsed=0;
  fat_item.Offset=-1;
  try
  {
    int i,j;
    int32_t val_db=IS_DB;
    CRC32 c;
    seekf(&db,0);
    writef(&db,&val_db,sizeof(int32_t));
    seekf(&db,GetOffset_Header());
    writef(&db,&db.header,sizeof(Header));
    for (i=0;i<db.header.NbLT;i++)
    {
      seekf(&db,GetOffset_LT_Begin(&db,i));
      for (j=0;j<nbItems;j++)
        writef(&db,&lt_item,sizeof(lt_item));
      Calc_LT_CRC(&db,i,c);
      Write_LT_CRC(&db,i,c);
    }
    for (i=0;i<db.header.NbFAT;i++)
    {
      seekf(&db,GetOffset_FAT_NbItems(&db,i));
      writef(&db,&fat_nbitems,sizeof(FAT_NbItems));
      seekf(&db,GetOffset_FAT_Begin(&db,i));
      for (j=0;j<nbItems;j++)
        writef(&db,&fat_item,sizeof(fat_item));
      Calc_FAT_CRC(&db,i,c);
      Write_FAT_CRC(&db,i,c);
    }
  }
  catch (int err_num)
  {
    if(err_num!=ERROR_SUCCESS)
    {
      fclose(db.fic);
      throw err_num;
    }
  }
  closef(db.fic);
}

DBASE OpenDB(char* DB)
{
  DBase* db=new (std::nothrow) DBase;
  if (db==NULL)
    throw ERROR_COULD_NOT_ALLOCATE_MEMORY;
  if (fopen_s(&(db->fic),DB,"rb+"))
    throw ERROR_COULD_NOT_OPEN_FILE;
  CRC32 (*lt_crc)[2]=NULL,(*fat_crc)[2]=NULL;

  try   //Gestion des erreurs : le cas chant, referme le fichier et renvoie le code d'erreur.
  {
    seekf(db,GetOffset_Header());
    readf(db,&db->header,sizeof(Header));

    lt_crc=new (std::nothrow) CRC32[db->header.NbLT][2];
    if (lt_crc==NULL)
      throw ERROR_COULD_NOT_ALLOCATE_MEMORY;
    fat_crc=new (std::nothrow) CRC32[db->header.NbFAT][2];
    if (fat_crc==NULL)
      throw ERROR_COULD_NOT_ALLOCATE_MEMORY;

    int i;
    for (i=0;i<db->header.NbLT;i++)
    {
      Read_LT_CRC(db,i,lt_crc[i][0]);
      Calc_LT_CRC(db,i,lt_crc[i][1]);

      //Aprs faire un truc plus complexe, c'est  dire lire une table qui ne serait pas endommag etc.
      if (lt_crc[i][0]!=lt_crc[i][1])
        throw ERROR_LT_CORRUPTED;
    }
    for (i=0;i<db->header.NbFAT;i++)
    {
      Read_FAT_CRC(db,i,fat_crc[i][0]);
      Calc_FAT_CRC(db,i,fat_crc[i][1]);

      //Aprs faire un truc plus complexe, c'est  dire lire une table qui ne serait pas endommag etc.
      if (fat_crc[i][0]!=fat_crc[i][1])
        throw ERROR_FAT_CORRUPTED;
    }

  db->lt=Read_LT(db,0);
  db->fat=Read_FAT(db,0);
  }
  catch (int err_num)
  {
    if(err_num!=ERROR_SUCCESS)
    {
      if (lt_crc!=NULL)   delete[] lt_crc;
      if (fat_crc!=NULL)  delete[] fat_crc;
      fclose(db->fic);
      throw err_num;
    }
  }
   

  delete[] lt_crc;
  delete[] fat_crc;
  return (DBASE)db;
}

void UpdateDB(DBASE dbase)
{
  DBase* db=(DBase*)dbase;
  int i;
  for (i=0;i<db->header.NbLT;i++)
    Write_LT(db,i);
  for (i=0;i<db->header.NbFAT;i++)
    Write_FAT(db,i);
}

void CloseDB(DBASE dbase)
{
  DBase* db=(DBase*)dbase;
  UpdateDB(dbase);
  delete[] db->lt.lt_item;
  delete[] db->fat.fat_item;
  closef(db->fic);
  delete db;
}

void DeleteDB(char* DB)
{
  if (remove(DB)!=0)
    throw ERROR_COULD_NOT_DELETE_FILE;
}

bool RecordExist(DBASE dbase,long n)
{
  DBase* db=(DBase*)dbase;
  if (n>=db->header.NbMaxItem)
    throw ERROR_RECORD_OUT_OF_BOUND;
  LT_Item lt_i;
  FAT_Item fat_i;
  lt_i=db->lt.lt_item[n];
  if (lt_i.FatOffset==-1)
    return false;
  fat_i=db->fat.fat_item[lt_i.FatOffset];
  if (fat_i.MemAlloc==0||fat_i.Offset==-1)
    return false;
  return true;
}

void CreateRecord(DBASE dbase,long n,long size)
{
  DBase* db=(DBase*)dbase;
  if (n>=db->header.NbMaxItem)
    throw ERROR_RECORD_OUT_OF_BOUND;
  if (RecordExist(dbase,n))
    throw ERROR_RECORD_ALREADY_EXIST;
  couple pos=FAT_Find_Mem(db,size);
  FAT_Push_Item(db,pos.a);
  LT_Push_Item(db,pos.a);
  db->fat.fat_nbitems++;
  db->fat.fat_item[pos.a].Offset=pos.b;
  db->fat.fat_item[pos.a].MemAlloc=size;
  db->fat.fat_item[pos.a].MemUsed=0;
  db->lt.lt_item[n].FatOffset=pos.a;
}

void DeleteRecord(DBASE dbase,long n)
{
  DBase* db=(DBase*)dbase;
  if (!RecordExist(dbase,n))
    throw ERROR_RECORD_DOES_NOT_EXIST;
  FAT_Pop_Item(db,db->lt.lt_item[n].FatOffset);
  LT_Pop_Item(db,db->lt.lt_item[n].FatOffset);
  db->fat.fat_nbitems--;
  db->lt.lt_item[n].FatOffset=-1;
}

RECORD OpenRecord(DBASE dbase,long n,int open_mode)
{
  DBase* db=(DBase*)dbase;
  if (!RecordExist(dbase,n))
    throw ERROR_RECORD_DOES_NOT_EXIST;
  Record *rec=new Record;
  rec->mode=Int_To_OpenMode(open_mode);
  rec->nb=n;
  rec->Alloc=Read_Record_MemAlloc(db,n);
  rec->db=db;
  if (rec->mode==DELETE)
  {
    rec->Used=0;
    FAT_Write_Record_MemUsed(db,n,0);
  }
  else
    rec->Used=Read_Record_MemUsed(db,n);
  if (rec->mode==APPEND)
    rec->Pos=Read_Record_MemUsed(db,n);
  else
    rec->Pos=0;
  return (RECORD)rec;
}

void WriteRecord(RECORD record,const void* buffer,long size)
{
  Record *rec=(Record*)record;
  if (rec->mode==READ)
    throw ERROR_WRONG_OPEN_MODE;
  if ((rec->Pos+size)>(rec->Alloc))
    throw ERROR_RECORD_OUT_OF_MEMORY;
  long off=GetOffset_DataBase(rec->db)+Read_Record_Offset(rec->db,rec->nb)+rec->Pos;
  seekf(rec->db,off);
  writef(rec->db,buffer,size);
  rec->Pos+=size;
  if (rec->Pos>rec->Used)
    rec->Used=rec->Pos;
}

void ReadRecord(RECORD record,void* buffer,long size)
{
  Record *rec=(Record*)record;
  if (rec->mode!=READ)
    throw ERROR_WRONG_OPEN_MODE;
  if ((rec->Pos+size)>(rec->Used))
    throw ERROR_RECORD_OUT_OF_MEMORY;
  long off=GetOffset_DataBase(rec->db)+Read_Record_Offset(rec->db,rec->nb)+rec->Pos;
  seekf(rec->db,off);
  readf(rec->db,buffer,size);
  rec->Pos+=size;
}

void SeekRecord(RECORD record,long offset,int mode)
{
  Record *rec=(Record*)record;
  long off;
  if (mode==MODE_ABS)
    off=offset;
  else if (mode==MODE_REL)
    off=rec->Pos+offset;
  else 
    throw ERROR_INVALIDE_MODE;
  if (off<0 || (rec->mode==READ && off>=rec->Used) || (rec->mode!=READ && off>=rec->Alloc))
      throw ERROR_SEEKING_OUT_OF_RECORD;
  
  rec->Pos=off;
}

void CloseRecord(RECORD record)
{
  Record *rec=(Record*)record;
  if (rec->mode!=READ)
  {
    long used=rec->Used;
    if (Read_Record_MemUsed(rec->db,rec->nb)>used)
      used=Read_Record_MemUsed(rec->db,rec->nb);
    FAT_Write_Record_MemUsed(rec->db,rec->nb,used);
  }
  delete rec;
}

void ResizeRecord(DBASE dbase,long n,long size)
{
  DBase* db=(DBase*)dbase;
  long mem_used=Read_Record_MemUsed(db,n);
  long mem_alloc=Read_Record_MemAlloc(db,n);
  if (size==mem_alloc)
    return;
  if (mem_used>size)
    throw ERROR_SIZE_SMALLER_THAN_RECORD;
  if (IsAvailable_Mem_After_Record(db,n,size-mem_alloc))
    FAT_Write_Record_MemAlloc(db,n,size);
  else
  {
    char *temp=new (std::nothrow) char[mem_used];
    if (temp!=NULL)
    {
      RECORD rec;
      rec=OpenRecord(dbase,n,MODE_READ);
      ReadRecord(rec,temp,mem_used);
      CloseRecord(rec);
      DeleteRecord(dbase,n);
      CreateRecord(dbase,n,size);
      rec=OpenRecord(dbase,n,MODE_WRITE);
      WriteRecord(rec,temp,mem_used);
      CloseRecord(rec);
      delete temp;
    }
    else
    {
      char t;
      int n_t=GetFirstEmptyRecord(dbase),i;
      RECORD from,to;
      CreateRecord(dbase,n_t,size);
      to=OpenRecord(dbase,n_t,MODE_WRITE);
      from=OpenRecord(dbase,n,MODE_READ);
      for (i=0;i<mem_used;i++)
      {
        ReadRecord(from,&t,1);
        WriteRecord(to,&t,1);
      }
      CloseRecord(from);
      CloseRecord(to);
      DeleteRecord(dbase,n);
      LT_Write_FAT_Offset(db,n,LT_Read_FAT_Offset(db,n_t));
      LT_Write_FAT_Offset(db,n_t,-1);
    }
  }
}

record_info GetRecordInfo(RECORD record)
{
  Record *rec=(Record*)record;
  record_info info;
  info.nb=rec->nb;
  info.open_mode=OpenMode_To_Int(rec->mode);
  info.MemAlloc=rec->Alloc;
  info.MemUsed=(rec->mode==READ)?rec->Used:rec->Pos;
  info.CurrentPos=rec->Pos;
  return info;
}

void UpdateRecord(RECORD record)
{
  Record *rec=(Record*)record;
  rec->Alloc=Read_Record_MemAlloc(rec->db,rec->nb);
  if (rec->mode==READ)
    rec->Used=Read_Record_MemUsed(rec->db,rec->nb);
  else
  {
    long used=Read_Record_MemUsed(rec->db,rec->nb);
    if (rec->Used<used)
      rec->Used=used;
    else
      FAT_Write_Record_MemUsed(rec->db,rec->nb,rec->Used);
  }
}

int GetNumberOfEmptyRecord(DBASE dbase)
{
  DBase* db=(DBase*)dbase;
  return db->header.NbMaxItem-db->fat.fat_nbitems;
}

int GetFirstEmptyRecord(DBASE dbase)
{
  DBase* db=(DBase*)dbase;
  int i;
  if (GetNumberOfEmptyRecord(dbase)==0)
    throw ERROR_NO_EMPTY_RECORD;
  for (i=0;i<db->header.NbMaxItem;i++)
    if (!RecordExist(dbase,i))
      return i;
  throw ERROR_NO_EMPTY_RECORD;
}

////////////////////////////
//// FONCTIONS DIVERSES ////
////////////////////////////

OpenMode Int_To_OpenMode(int i)
{
  switch(i)
  {
    case MODE_READ:
      return READ;
    case MODE_WRITE:
      return WRITE;
    case MODE_APPEND:
      return APPEND;
    case MODE_DELETE:
      return DELETE;
    default:
      throw ERROR_INVALIDE_MODE;
  }
}

int OpenMode_To_Int(OpenMode open_mode)
{
  switch (open_mode)
  {
    case READ:
      return MODE_READ;
    case WRITE:
      return MODE_WRITE;
    case APPEND:
      return MODE_APPEND;
    case DELETE:
      return MODE_DELETE;
    default:
      return ABRUTI_DE_COMPILO;
  }
}
/*
int truncate(long size)
{
#ifdef _MSC_VER
  seekf(size);
  SetEndOfFile(db.header.
#else
#endif
}*/
//////////////////////////////////////////////
//// FONCTIONS DE MANIPULATION DES TABLES ////
//////////////////////////////////////////////

couple FAT_Find_Mem(DBase* db,long size)
{
  int i=0;
  couple r;
  //S'il n'y a pas d'lment dans la table ou que le premier doit-tre repouss.
  if ((db->fat.fat_nbitems==0)||(db->fat.fat_item[0].Offset>=size))
  {
    r.a=0;
    r.b=0;
    return r;
  }
  else
  {
    for (i=1;i<db->fat.fat_nbitems;i++)
      if ((db->fat.fat_item[i].Offset-(db->fat.fat_item[i-1].Offset+db->fat.fat_item[i-1].MemAlloc))>=size)
        break;
    r.a=i;
    r.b=db->fat.fat_item[i-1].Offset+db->fat.fat_item[i-1].MemAlloc;
    return r;
  }
}

bool IsAvailable_Mem_After_Record(DBase* db,long nb_record,long size)
{
  long ordre=LT_Read_FAT_Offset(db,nb_record);
  if (ordre>=db->fat.fat_nbitems)
    throw ERROR_RECORD_DOES_NOT_EXIST;
  if (ordre==db->fat.fat_nbitems)
    return true;
  else
    if ((FAT_Read_Record_Offset(db,ordre+1)-(FAT_Read_Record_Offset(db,ordre)+FAT_Read_Record_MemAlloc(db,ordre)))>=size)
      return true;
    else
      return false;
}

void LT_Push_Item(DBase* db,long n_start)
{
  int i;
  for (i=0;i<db->header.NbMaxItem;i++)
    if (db->lt.lt_item[i].FatOffset>n_start)
      db->lt.lt_item[i].FatOffset++;
}

void LT_Pop_Item(DBase* db,long n_start)
{
  int i;
  for (i=0;i<db->header.NbMaxItem;i++)
    if (db->lt.lt_item[i].FatOffset>n_start)
      db->lt.lt_item[i].FatOffset--;
}

void FAT_Push_Item(DBase* db,long n_start)
{
  int i;
  for (i=db->fat.fat_nbitems;i>n_start;i--)
    db->fat.fat_item[i]=db->fat.fat_item[i-1];
}

void FAT_Pop_Item(DBase* db,long n_start)
{
  int i;
  for (i=n_start;i<db->fat.fat_nbitems-1;i++)
    db->fat.fat_item[i]=db->fat.fat_item[i+1];
  db->fat.fat_item[i].MemAlloc=0;
  db->fat.fat_item[i].MemUsed=0;
  db->fat.fat_item[i].Offset=-1;
}

long LT_Read_FAT_Offset(DBase* db,int nb_record)
{
  if (nb_record>=db->header.NbMaxItem)
    throw ERROR_RECORD_OUT_OF_BOUND;
  if (db->lt.lt_item[nb_record].FatOffset==-1)
    throw ERROR_RECORD_DOES_NOT_EXIST;
  return db->lt.lt_item[nb_record].FatOffset;
}

void LT_Write_FAT_Offset(DBase* db,int nb_record,long fat_offset)
{
  if (nb_record>=db->header.NbMaxItem)
    throw ERROR_RECORD_OUT_OF_BOUND;
  db->lt.lt_item[nb_record].FatOffset=fat_offset;
}

long FAT_Read_Record_Offset(DBase* db,int ordre_record)
{
  if (ordre_record>=db->fat.fat_nbitems)
    throw ERROR_RECORD_DOES_NOT_EXIST;
  return db->fat.fat_item[ordre_record].Offset;
}

long FAT_Read_Record_MemAlloc(DBase* db,int ordre_record)
{
  if (ordre_record>=db->fat.fat_nbitems)
    throw ERROR_RECORD_DOES_NOT_EXIST;
  return db->fat.fat_item[ordre_record].MemAlloc;
}

long FAT_Read_Record_MemUsed(DBase* db,int ordre_record)
{
  if (ordre_record>=db->fat.fat_nbitems)
    throw ERROR_RECORD_DOES_NOT_EXIST;
  return db->fat.fat_item[ordre_record].MemUsed;
}

long Read_Record_Offset(DBase* db,int nb_record)
{
  return FAT_Read_Record_Offset(db,LT_Read_FAT_Offset(db,nb_record));
}

long Read_Record_MemAlloc(DBase* db,int nb_record)
{
  return FAT_Read_Record_MemAlloc(db,LT_Read_FAT_Offset(db,nb_record));
}

long Read_Record_MemUsed(DBase* db,int nb_record)
{
  return FAT_Read_Record_MemUsed(db,LT_Read_FAT_Offset(db,nb_record));
}

void FAT_Write_Record_MemUsed(DBase* db,int nb_record,long MemUsed)
{
  long nb=LT_Read_FAT_Offset(db,nb_record);
  if (nb>=db->fat.fat_nbitems)
    throw ERROR_RECORD_DOES_NOT_EXIST;
  db->fat.fat_item[nb].MemUsed=MemUsed;
}
void FAT_Write_Record_MemAlloc(DBase* db,int nb_record,long MemAlloc)
{
  long nb=LT_Read_FAT_Offset(db,nb_record);
  if (nb>=db->fat.fat_nbitems)
    throw ERROR_RECORD_DOES_NOT_EXIST;
  db->fat.fat_item[nb].MemAlloc=MemAlloc;
}

//////////////////////////////////////////////
//// FONCTIONS D'ECRITURE DANS LES TABLES ////
//////////////////////////////////////////////

void Write_LT(DBase* db,int n)
{
  seekf(db,GetOffset_LT_Begin(db,n));
  int i;
  CRC32 c;
  for (i=0;i<db->header.NbMaxItem;i++)
    writef(db,&db->lt.lt_item[i],sizeof(LT_Item));
  Calc_LT_CRC(db,n,c);
  seekf(db,GetOffset_LT_CRC(db,n));
  writef(db,&c,sizeof(CRC32));
}

void Write_FAT(DBase* db,int n)
{
  int i;
  CRC32 c;
  seekf(db,GetOffset_FAT_NbItems(db,n));
  writef(db,&db->fat.fat_nbitems,sizeof(FAT_NbItems));
  seekf(db,GetOffset_FAT_Begin(db,n));
  for (i=0;i<db->fat.fat_nbitems;i++)
    writef(db,&db->fat.fat_item[i],sizeof(FAT_Item));
  Calc_FAT_CRC(db,n,c);
  seekf(db,GetOffset_FAT_CRC(db,n));
  writef(db,&c,sizeof(CRC32));
}

//////////////////////////////////////////////
//// FONCTIONS DE LECTURE DANS LES TABLES ////
//////////////////////////////////////////////

LT_Item LT_Read_Record_Data(DBase* db,int nb_lt,int nb_record)
{
  if (nb_lt<db->header.NbLT)
    if (nb_record<db->header.NbMaxItem)
      {
        LT_Item item;
        seekf(db,GetOffset_LT_Begin(db,nb_lt)+(long)sizeof(LT_Item)*nb_record);
        readf(db,&item,sizeof(LT_Item));
        return item;
      }
    else
      throw ERROR_RECORD_OUT_OF_BOUND;
  else
    throw ERROR_LT_DOES_NOT_EXIST;
}

long FAT_Read_NbItems(DBase* db,int nb_fat)
{
  if (nb_fat<db->header.NbFAT)
  {
    seekf(db,GetOffset_FAT_NbItems(db,nb_fat));
    FAT_NbItems nb;
    readf(db,&nb,sizeof(FAT_NbItems));
    return nb;
  }
  else
    throw  ERROR_FAT_DOES_NOT_EXIST;
}

FAT_Item FAT_Read_Record_Data(DBase* db,int nb_fat,int ordre_record)
{
  if (nb_fat<db->header.NbFAT)
    if (ordre_record<FAT_Read_NbItems(db,nb_fat))
    {
      FAT_Item item;
      seekf(db,GetOffset_FAT_Begin(db,nb_fat)+(long)sizeof(FAT_Item)*ordre_record);
      readf(db,&item,sizeof(FAT_Item));
      return item;
    }
    else
      throw ERROR_RECORD_DOES_NOT_EXIST;
  else
    throw ERROR_FAT_DOES_NOT_EXIST;
}

LT Read_LT(DBase* db,int n)
{
  LT l;
  int i;
  l.lt_item=new (std::nothrow) LT_Item[db->header.NbMaxItem];
  if (l.lt_item==NULL)
      throw ERROR_COULD_NOT_ALLOCATE_MEMORY;
  try
  {
    for (i=0;i<db->header.NbMaxItem;i++)
      l.lt_item[i]=LT_Read_Record_Data(db,n,i);
  }
  catch (int err_num)
  {
    if (err_num!=ERROR_SUCCESS)
      delete[] l.lt_item;
    throw err_num;
  }
  return l;
}

FAT Read_FAT(DBase* db,int n)
{
  FAT f;
  int i;
  f.fat_item=new (std::nothrow) FAT_Item[db->header.NbMaxItem];
  if (f.fat_item==NULL)
      throw ERROR_COULD_NOT_ALLOCATE_MEMORY;
  try
  {
    f.fat_nbitems=FAT_Read_NbItems(db,n);
    for (i=0;i<f.fat_nbitems;i++)
      f.fat_item[i]=FAT_Read_Record_Data(db,n,i);
  }
  catch (int err_num)
  {
    if (err_num!=ERROR_SUCCESS)
      delete[] f.fat_item;
    throw err_num;
  }
  return f;
}

//////////////////////////////////////////
//// FONCTIONS DE MANIPULATION DE CRC ////
//////////////////////////////////////////

void Calc_CRC(DBase* db,long begin,long size,CRC32 &c)
{
  int i;
  char l=0;
  c=0;
  seekf(db,begin);
  for (i=0;i<size;i++)
  {
    readf(db,&l,1);
    c|=(l<<(8*(i%4)));
  }
}

void Calc_LT_CRC(DBase* db,int n,CRC32 &c)
{
  if (n<db->header.NbLT)
    Calc_CRC(db,GetOffset_LT_Begin(db,n),sizeof(LT_Item)*db->header.NbMaxItem,c);
  else
    throw ERROR_LT_DOES_NOT_EXIST;
}

void Calc_FAT_CRC(DBase* db,int n,CRC32 &c)
{
  if (n<db->header.NbFAT)
    Calc_CRC(db,GetOffset_FAT_Begin(db,n),sizeof(FAT_Item)*db->header.NbMaxItem,c);
  else
    throw ERROR_FAT_DOES_NOT_EXIST;
}

void Read_CRC(DBase* db,long Offset,CRC32 &c)
{
  seekf(db,Offset);
  readf(db,&c,sizeof(CRC32));
}

void Read_LT_CRC(DBase* db,int n,CRC32 &c)
{
  if (n<db->header.NbLT)
    Read_CRC(db,GetOffset_LT_CRC(db,n),c);
  else
    throw ERROR_LT_DOES_NOT_EXIST;
}

void Read_FAT_CRC(DBase* db,int n,CRC32 &c)
{
  if (n<db->header.NbFAT)
    Read_CRC(db,GetOffset_FAT_CRC(db,n),c);
  else
    throw ERROR_FAT_DOES_NOT_EXIST;
}

void Write_CRC(DBase* db,long Offset,CRC32 &c)
{
  seekf(db,Offset);
  writef(db,&c,sizeof(CRC32));
}

void Write_LT_CRC(DBase* db,int n,CRC32 &c)
{
  if (n<db->header.NbLT)
    Write_CRC(db,GetOffset_LT_CRC(db,n),c);
  else
    throw ERROR_LT_DOES_NOT_EXIST;

}

void Write_FAT_CRC(DBase* db,int n,CRC32 &c)
{
  if (n<db->header.NbFAT)
    Write_CRC(db,GetOffset_FAT_CRC(db,n),c);
  else
    throw ERROR_FAT_DOES_NOT_EXIST;
}

////////////////////////////////////////
//// FONCTIONS DE CALCUL D'ADRESSES ////
////////////////////////////////////////

long GetOffset_Header()
{
  return (long)(sizeof(int32_t));
}

long GetOffset_First_LT()
{
  return GetOffset_Header()+(long)sizeof(Header);
}

long GetOffset_LT_CRC(DBase* db,int n)
{
  if (n<db->header.NbLT)
    return GetOffset_First_LT()+(long)((sizeof(CRC32)+sizeof(LT_Item)*db->header.NbMaxItem)*n);
  else
    throw ERROR_LT_DOES_NOT_EXIST;
}

long GetOffset_LT_Begin(DBase* db,int n)
{
  if (n<db->header.NbLT)
    return GetOffset_LT_CRC(db,n)+(long)(sizeof(CRC32));
  else
    throw ERROR_LT_DOES_NOT_EXIST;
}

long GetOffset_First_FAT(DBase* db)
{
  return GetOffset_First_LT()+(long)((sizeof(CRC32)+sizeof(LT_Item)*db->header.NbMaxItem)*db->header.NbLT);
}

long GetOffset_FAT_CRC(DBase* db,int n)
{
  if (n<db->header.NbFAT)
    return GetOffset_First_FAT(db)+(long)((sizeof(CRC32)+sizeof(long)+sizeof(FAT_Item)*db->header.NbMaxItem)*n);
  else
    throw ERROR_FAT_DOES_NOT_EXIST;
}

long GetOffset_FAT_NbItems(DBase* db,int n)
{
  if (n<db->header.NbFAT)
    return GetOffset_FAT_CRC(db,n)+(long)(sizeof(CRC32));
  else
    throw ERROR_FAT_DOES_NOT_EXIST;
}

long GetOffset_FAT_Begin(DBase* db,int n)
{
  if (n<db->header.NbFAT)
    return GetOffset_FAT_NbItems(db,n)+(long)(sizeof(FAT_NbItems));
  else
    throw ERROR_FAT_DOES_NOT_EXIST;
}

#ifdef LEGACY_MODE
long GetOffset_DataBase(DBase* db)
{
  return (long)(sizeof(Header)+db->header.NbLT*(sizeof(CRC32)+db->header.NbMaxItem*sizeof(LT_Item))+db->header.NbFAT*(sizeof(CRC32)+db->header.NbMaxItem*sizeof(FAT_Item)));
}
#else
long GetOffset_DataBase(DBase* db)
{
  return (long)(sizeof(int32_t)+sizeof(Header)+db->header.NbLT*(sizeof(CRC32)+db->header.NbMaxItem*sizeof(LT_Item))+db->header.NbFAT*(sizeof(CRC32)+db->header.NbMaxItem*sizeof(FAT_Item)));
}
#endif

static char* assign_error(char* str)
{
  int i=0;
  while(str[i]!=0)
  {
    error[i]=str[i];
    i++;
  }
  error[i]=0;
  return error;
}

char* GetErrorString(int error)
{
  switch (error)
  {
/* 0 */  case ERROR_SUCCESS:
           return assign_error("ERREUR : Pas d'erreur.");
/* 1 */  case ERROR_COULD_NOT_OPEN_FILE:
           return assign_error("ERREUR : Le fichier n'a pas pu tre ouvert (il n'existe pas, ou il est utilis par un autre programme).");
/* 2 */  case ERROR_READ_ERROR:
           return assign_error("ERREUR : Erreur en lecture.");
/* 3 */  case ERROR_WRITE_ERROR:
           return assign_error("ERREUR : Erreur en criture.");
/* 4 */  case ERROR_SEEK_ERROR:
           return assign_error("ERREUR : Erreur de positionnement dans le fichier.");
/* 5 */  case ERROR_LT_CORRUPTED :
           return assign_error("ERREUR : La table de lien est corrompu.");
/* 6 */  case ERROR_FAT_CORRUPTED:
           return assign_error("ERREUR : La table d'allocation est corrompu.");
/* 7 */  case ERROR_LT_DOES_NOT_EXIST:
           return assign_error("ERREUR : La table de lien n'existe pas.");
/* 8 */  case ERROR_FAT_DOES_NOT_EXIST:
           return assign_error("ERREUR : La table d'allocation n'existe pas.");
/* 9 */  case ERROR_RECORD_DOES_NOT_EXIST:
           return assign_error("ERREUR : L'enregistrement demand n'existe pas dans la table d'allocation (il n'a pas t cr).");
/* 10 */ case ERROR_RECORD_OUT_OF_BOUND:
           return assign_error("ERREUR : L'enregistrement demand a un numro d'ordre suprieur a NbMaxItem (utilis pendant la cration de la table).");
/* 11 */ case ERROR_COULD_NOT_ALLOCATE_MEMORY:
           return assign_error("ERREUR : Une allocation de mmoire  chou.");
/* 12 */ case ERROR_DB_ALREADY_OPENED:
           return assign_error("ERREUR : Une base de donns est dj ouverte.");
/* 13 */ case ERROR_COULD_NOT_CREATE_FILE:
           return assign_error("ERREUR : Le fichier n'a pas pu tre cr.");
/* 14 */ case ERROR_FILE_ALREADY_EXIST:
           return assign_error("ERREUR : Le fichier existe dj.");
/* 15 */ case ERROR_DATABASE_IS_NOT_OPEN:
           return assign_error("ERREUR : Aucune base de donnes n'est ouverte.");
/* 16 */ case ERROR_COULD_NOT_DELETE_FILE:
           return assign_error("ERREUR : Le fichier n'a pas pu tre suprim (il est utilis par une aplication, ou il n'existe pas).");
/* 17 */ case ERROR_RECORD_ALREADY_EXIST:
           return assign_error("ERREUR : L'enregistrement que vous voulez crer existe dj.");
/* 18 */ case ERROR_CLOSE_ERROR:
           return assign_error("ERREUR : Erreur lors de la fermeture du fichier.");
/* 19 */ case ERROR_INVALIDE_MODE:
           return assign_error("ERREUR : Le mode demand n'existe pas.");
/* 20 */ case ERROR_RECORD_OUT_OF_MEMORY:
           return assign_error("ERREUR : Vous avez dpass la fin de la mmoire allou pour l'enregistrement (en criture), ou la fin de la mmoire utilis (en lecture).");
/* 21 */ case ERROR_WRONG_OPEN_MODE:
           return assign_error("ERREUR : L'opration demand n'est pas conforme au mode d'ouverture de l'enregistrement.");
/* 22 */ case ERROR_NO_EMPTY_RECORD:
           return assign_error("ERREUR : Tout les enregistrement sont utilis.");
/* 23 */ case ERROR_SIZE_SMALLER_THAN_RECORD:
           return assign_error("ERREUR : La taille demand est plus petite que celle utilis actuellement par l'enregistrement.");
/* 24 */ case ERROR_SEEKING_OUT_OF_RECORD:
           return assign_error("ERREUR : Vous avez donnez une adresse  la fonction SeekRecord qui est au del de la mmoire dj crite (en mode lecture) ou au del de la mmoire allou (en mode criture).");
/* 25 */ case ERROR_UNKNOWN_ERROR:
           return assign_error("ERREUR : Aucune erreur avec ce numro n'existe.");
         default:
           throw ERROR_UNKNOWN_ERROR;
  }

}

