/*****************************************************************************
* Copyright (C) 2006, Mathias Kende                                          *
* licence@sectionpc.info                                                     *
*                                                                            *
* Le prsent logiciel est distribu sous la licence SectionPC dont vous avez *
* du avoir une copie.  Si ce n'est pas le cas,  vous pouvez vous en procurer *
* une  l'adresse http://licence.sectionpc.info  ou par courriel  l'adresse *
* ci-dessus.                                                                 *
*****************************************************************************/

#ifndef _TPBD_
  #define _TPBD_


#ifdef _MSC_VER   //Sur VC++
  #define USING_VCPP
#else             //Sur Dev C++
  #define USING_DEVCPP
#endif


#ifndef TPBD_COMPIL    //C'est une constante dfinit lors de la compilation de la lib elle mme.
#ifdef _DLL
#pragma comment(lib,"TPBD_dll")
#else
#pragma comment(lib,"TPBD")
#endif
#endif

#ifdef ERROR_SUCCESS
#   undef ERROR_SUCCESS
#endif

//////////////////////////////////////////////////
//// FONCTIONS D'INTERFACE AVEC L'UTILISATEUR ////
//////////////////////////////////////////////////

typedef void* RECORD;
typedef void* DBASE;

struct record_info
{
  long nb;            //Le numro de l'enregistrement.
  long open_mode;     //Le mode d'ouverture.
  long MemAlloc;      //La mmoire alloue pour cet enregistrement.
  long MemUsed;       //La mmoire utilise par l'enregistrement.
  long CurrentPos;    //La position du curseur d'criture ou de lecture dans l'enregistrement.
};

//Indique si le fichier est une base de donnes. Renvoie l'un des code de type FILE_IS... (voir plus bas).
int DBExist(char* DB);

//Cre une nouvelle base de donnes stock dans le fichier dont le nom est DB et pouvant contenir nbItems objets.
void CreateDB(char* DB,int nbItems);

//Ouvre une base de donnes existante.
DBASE OpenDB(char* DB);

//Mets  jour les LT et FAT de la DB. Cette opration n'est effectu que lors de CloseDB si vous ne le faites pas manuellement.
void UpdateDB(DBASE dbase);

//Ferme la base de donnes ouverte actuellement.
void CloseDB(DBASE dbase);

//Supprime une base de donnes.
void DeleteDB(char* DB);

//Compresse la base de donne en supprimant l'espace inutilis.
void CompressDB(int compression);  //Pour l'instant, a ne fait rien.

//Indique si l'enregistrement n existe ou non dans les FAT.
bool RecordExist(DBASE dbase,long n);

//Cre l'enregistrement n qui n'existe pas encore et lui alloue size octets.
void CreateRecord(DBASE dbase,long n,long mem_alloc);

//Supprime l'enregistrement n.
void DeleteRecord(DBASE dbase,long n);

//Ouvre l'enregistrement n en mode open_mode (voir les constantes ci-dessous).
RECORD OpenRecord(DBASE dbase,long n,int open_mode);

//Ecrit le contenu de buffer qui fait size octets dans l'enregistrement record.
void WriteRecord(RECORD record,const void* buffer,long size);

//Lis size octets de l'enregistrement record et les enregistre dans buffer.
void ReadRecord(RECORD record,void* buffer,long size);

//Modifie la position du pointeur de lecture/criture dans l'enregistrement.
void SeekRecord(RECORD record,long offset,int mode);

//Ferme un enregistrement ouvert.
void CloseRecord(RECORD record);

//Modifie la taille de l'enregistrement.
void ResizeRecord(DBASE dbase,long n,long mem_alloc);

//Renvoie une structure contenant des informations sur l'enregistrement record qui a t pralablement ouvert.
record_info GetRecordInfo(RECORD record);

//Mets  jours les informations d'un enregistrement qui aurait t modifi depuis son ouverture.
void UpdateRecord(RECORD record);

//Renvoie le nombre d'enregistrement libre dans la base de donnes.
int GetNumberOfEmptyRecord(DBASE dbase);

//Renvoie le numro du premier enregistrement qui n'est pas encore cr.
int GetFirstEmptyRecord(DBASE dbase);

char* GetErrorString(int error);

//Les code de retour de la fonction DBExist
#define FILE_DOES_NOT_EXIST                0    /* Le fichier n'existe pas. */
#define FILE_IS_NOT_DB                     1    /* Le fichier existe mais n'est pas une base de donnes */
#define FILE_IS_DB                         2    /* Le fichier existe et est une base de donnes. */

//Les diffrents modes d'ouvertures d'un enregistrement.
#define MODE_READ                           0   /* Ouvre l'enregistrement en mode lecture seulement. */
#define MODE_WRITE                          1   /* Ouvre l'enregistrement en mode criture seulement. */
#define MODE_APPEND                         2   /* Ouvre l'enregistrement en mode criture seulement pour un ajout de donnes  la fin de l'enregistrement. */
#define MODE_DELETE                         3   /* Efface l'enregistrement et l'ouvre en mode criture seulement. */

//Les diffrents modes pour modifier la positions du pointeurs d'un enregistrement.
#define MODE_ABS                            0   /* La position indique est relative au dbut du fichier. */
#define MODE_REL                            1   /* La position indique est relative  la position courante du pointeur. */

//Les degrs de compression. (ne servent  rien, car la compression ne fonctionne pas encore.
#define LOW_COMPRESSION                     0   /* Supprime l'espace libre qu'il peut y avoir entre les enregistrements. */
#define NORMAL_COMPRESSION                  1   /* Supprime l'espace libre qu'il peut y avoir entre les enregistrement et l'espace inutilis par les enregistremet. */

//Constante dfinissant les codes d'erreurs qui peuvent tre renvoy lors d'une exception.
#define ERROR_SUCCESS                       0   /* #Pas d'erreur */
#define ERROR_COULD_NOT_OPEN_FILE           1   /* Le fichier n'a pas pu tre ouvert (il n'existe pas, ou il est utilis par un autre programme). */
#define ERROR_READ_ERROR                    2   /* #Erreur en lecture */
#define ERROR_WRITE_ERROR                   3   /* #Erreur en criture */
#define ERROR_SEEK_ERROR                    4   /* #Erreur de positionnement dans le fichier */
#define ERROR_LT_CORRUPTED                  5   /* #La table de lien est corrompu */
#define ERROR_FAT_CORRUPTED                 6   /* #La table d'allocation est corrompu */
#define ERROR_LT_DOES_NOT_EXIST             7   /* #La table de lien n'existe pas */
#define ERROR_FAT_DOES_NOT_EXIST            8   /* #La table d'allocation n'existe pas */
#define ERROR_RECORD_DOES_NOT_EXIST         9   /* L'enregistrement demand n'existe pas dans la table d'allocation (il n'a pas t cr) */
#define ERROR_RECORD_OUT_OF_BOUND           10  /* L'enregistrement demand a un numro d'ordre suprieur a NbMaxItem (utilis pendant la cration de la table)*/
#define ERROR_COULD_NOT_ALLOCATE_MEMORY     11  /* #Une allocation de mmoire  chou */
#define ERROR_DB_ALREADY_OPENED             12  /* Une base de donns est dj ouverte. */
#define ERROR_COULD_NOT_CREATE_FILE         13  /* #Le fichier n'a pas pu tre cr. */
#define ERROR_FILE_ALREADY_EXIST            14  /* Le fichier existe dj. */
#define ERROR_DATABASE_IS_NOT_OPEN          15  /* #Aucune base de donnes n'est ouverte. Cette erreur est obsolte avec cette version. */
#define ERROR_COULD_NOT_DELETE_FILE         16  /* Le fichier n'a pas pu tre suprim (il est utilis par une aplication, ou il n'existe pas). */
#define ERROR_RECORD_ALREADY_EXIST          17  /* L'enregistrement que vous voulez crer existe dj. */
#define ERROR_CLOSE_ERROR                   18  /* #Erreur lors de la fermeture du fichier. */
#define ERROR_INVALIDE_MODE                 19  /* Le mode demand n'existe pas. */
#define ERROR_RECORD_OUT_OF_MEMORY          20  /* Vous avez dpass la fin de la mmoire allou pour l'enregistrement (en criture), ou la fin de la mmoire utilis (en lecture). */
#define ERROR_WRONG_OPEN_MODE               21  /* L'opration demand n'est pas conforme au mode d'ouverture de l'enregistrement. */
#define ERROR_NO_EMPTY_RECORD               22  /* Tout les enregistrement sont utilis. */
#define ERROR_SIZE_SMALLER_THAN_RECORD      23  /* La taille demand est plus petite que celle utilis actuellement par l'enregistrement. */
#define ERROR_SEEKING_OUT_OF_RECORD         24  /* Vous avez donnez une adresse  la fonction SeekRecord qui est au del de la mmoire dj crite (en mode lecture) ou au del de la mmoire allou (en mode criture). */
#define ERROR_UNKNOWN_ERROR                 25  /* Aucune erreur avec ce numro n'existe. */

/*
  Les constantes dont la description dbute par un '#' indique que les erreurs en questions ne sont
  pas susceptible de se produire pendant une utilisation normale de la bibliothque (sauf erreur systme par exemple).
  Les autres ne devrait pas se produire si mon code ne contient pas d'erreur et si le votre non plus.
*/

#endif

