/*****************************************************************************
* Copyright (C) 2006, Mathias Kende                                          *
* licence@sectionpc.info                                                     *
*                                                                            *
* Le prsent logiciel est distribu sous la licence SectionPC dont vous avez *
* du avoir une copie.  Si ce n'est pas le cas,  vous pouvez vous en procurer *
* une  l'adresse http://licence.sectionpc.info  ou par courriel  l'adresse *
* ci-dessus.                                                                 *
*****************************************************************************/

//Pour visual C++
#ifdef _MSC_VER
//# include <windows.h>
  typedef __int32 int32_t;
#else
# include <stdint.h>
# include <unistd.h>
# define fopen_s(fil,str,mod) ((*fil=fopen(str,mod))==NULL)
#endif

#include <cstdio>
#include <new>

typedef int32_t CRC32;

//Struture de l'en tte d'un fichier de base de donnes.
struct Header
{
  CRC32 CRC;                //CRC de l'ensemble des tables.
  int32_t NbFAT;
  int32_t NbLT;
  int32_t NbMaxItem;           //Le nombre d'lment pour lequel les FAT et LT sont allou (ce qui permet de calculer leur taille et donc leur adresses).
};

//Chaque lement d'une LT est l'adresse relative au dbut de la FAT de l'enregistrement en question.
struct LT_Item
{
  int32_t FatOffset;
};

typedef int32_t FAT_NbItems;    //Le nombre d'lment enregistr dans une FAT.

struct FAT_Item
{
  int32_t Offset;              //L'adresse de l'lment relative au dbut de la base de donnes.
  int32_t MemAlloc;           //La mmoire allou pour l'enregistrement (on alloue au moins sizeof(CRC32) de plus que ce que veut l'utilisateur).
  int32_t MemUsed;            //La mmoire utilis par l'enregistrement (y compris le CRC).
};

struct LT
{
  CRC32 crc;
  LT_Item *lt_item;
};
struct FAT
{
  CRC32 crc;
  FAT_NbItems fat_nbitems;  //Combien d'lment ont actuellement enregistr dans la FAT
  FAT_Item *fat_item;       //Les entre de la FAT elle mme.
};

//La structure de la base de donnes elle mme.
struct DBase
{
  FILE* fic;
  Header header;
  LT lt;
  FAT fat;
};

//Une structure qui sers  renvoyer des couples de valeurs quand c'est ncessaire.
struct couple
{
  long a,b;               
};

enum OpenMode
{
  READ=MODE_READ,
  WRITE=MODE_WRITE,
  APPEND=MODE_APPEND,
  DELETE=MODE_DELETE
};

//La structure d'un enregistrement de la base de donnes.
struct Record
{
  DBase* db;
  long nb;                //Numro de l'enregistrement lu.
  OpenMode mode;
  long Pos;               //La position, relative au dbut de l'enregistrement o en est le pointeur de lecture ou d'ecriture.
  long Alloc;             //Le nombre d'octet allou  cet enregistrement.
  long Used;              //Le nombre d'octet que contient cet enregistrement au moment o il est ouvert
};
extern int res;
#define readf(db,pt,sz)  {int res=(int)fread(pt,sz,1,(*db).fic);       if (res<1) throw ERROR_READ_ERROR;  }
#define writef(db,pt,sz) {int res=(int)fwrite(pt,sz,1,(*db).fic);      if (res<1) throw ERROR_WRITE_ERROR; }
#define seekf(db,off)       {int res=(int)fseek((*db).fic,off,SEEK_SET);  if (res!=0) throw ERROR_SEEK_ERROR; }
#define closef(ff)          {int res=(int)fclose(ff);               if (res!=0) throw ERROR_CLOSE_ERROR;}
//#define OnError(var)  if(var!=ERROR_SUCCESS) {fclose(fic); return var;}

#define ABRUTI_DE_COMPILO   0
#define IS_DB               673123743L    /* Code arbitraire indiquant que le fichier est bien une base de donnes */

////////////////////////////
//// FONCTIONS DIVERSES ////
////////////////////////////

//Convertit un entier en un type d'numration OpenMode.
OpenMode Int_To_OpenMode(int i);

//Convertit un membre d'numaration OpenMode en entier.
int OpenMode_To_Int(OpenMode open_mode);

//////////////////////////////////////////////
//// FONCTIONS DE MANIPULATION DES TABLES ////
//////////////////////////////////////////////

//Trouve le premier emplacement dans le fichier de size octets libre conscutif et renvoie son offset et l'enregistrement suivant.
couple FAT_Find_Mem(DBase* db,long size);

//Indique si size octets sont libres aprs l'enregistrement nb_record et avant le suivant.
bool IsAvailable_Mem_After_Record(DBase* db,long nb_record,long size);

//Dcale vers le haut les indices des lments de la LT pour lesquels c'est ncessaire.
void LT_Push_Item(DBase* db,long n_start);

//Dcale vers le bas les indices de la LT pour tout les lments dont c'est ncessaire.
void LT_Pop_Item(DBase* db,long n_start);

//Dcale tout les lment de la FAT  partir du rang n_start de 1 cran vers la fin.
void FAT_Push_Item(DBase* db,long n_start);

//Supprime l'lment n_start de la FAT et compacte les lments suivant.
void FAT_Pop_Item(DBase* db,long n_start);

//Renvoie le numro d'ordre relative au dbut d'une FAT, de l'enregistrement concernant l'lment nb_record.
long LT_Read_FAT_Offset(DBase* db,int nb_record);

//Ecrit dans la table de lien, l'offset o se trouve les information concernant l'enregistrement nb_record dans la table d'allocation.
void LT_Write_FAT_Offset(DBase* db,int nb_record,long fat_offset);

//Renvoie l'adresse relative au dbut de la zone de donne de l'enregistrement d'ordre ordre_record dans la FAT.
long FAT_Read_Record_Offset(DBase* db,int ordre_record);

//Renvoie la mmoire allou  l'enregistrement d'ordre ordre_record dans la FAT.
long FAT_Read_Record_MemAlloc(DBase* db,int ordre_record);

//Renvoie la mmoire utilis par l'enregistrement d'ordre ordre_record dans la FAT.
long FAT_Read_Record_MemUsed(DBase* db,int ordre_record);

//Renvoie l'adresse relative au dbut de la zone de donne de l'enregistrement de numro nb_record.
long Read_Record_Offset(DBase* db,int nb_record);

//Renvoie la mmoire allou pour l'enregistrement de numro nb_record.
long Read_Record_MemAlloc(DBase* db,int nb_record);

//Renvoie la mmoire utilis par l'enregistrement de numro nb_record.
long Read_Record_MemUsed(DBase* db,int nb_record);

void FAT_Write_Record_MemUsed(DBase* db,int nb_record,long MemUsed);

void FAT_Write_Record_MemAlloc(DBase* db,int nb_record,long MemAlloc);

//////////////////////////////////////////////
//// FONCTIONS D'ECRITURE DANS LES TABLES ////
//////////////////////////////////////////////

//Ecrit la LT n.
void Write_LT(DBase* db,int n);

//Ecrit la FAT n.
void Write_FAT(DBase* db,int n);

//////////////////////////////////////////////
//// FONCTIONS DE LECTURE DANS LES TABLES ////
//////////////////////////////////////////////

//Renvoie les donnes contenues dans la LT numro nb_lt  propos de l'enregistrement numro nb_record.
LT_Item LT_Read_Record_Data(DBase* db,int nb_lt,int nb_record);

//Renvoie le nombre d'enregistrement effectivemetn prsent dans la FAT nb_fat.
long FAT_Read_NbItems(DBase* db,int nb_fat);

//Renvoie les donnes concernant l'enregistrement d'ordre ordre_record dans la fat nb_fat.
FAT_Item FAT_Read_Record_Data(DBase* db,int nb_fat,int ordre_record);

//Lis une table de lien et alloue la mmoire pour en sauvegarder la structure (la mmoire est dsalloue dans CloseDB()).
LT Read_LT(DBase* db,int n);

//Lis une table d'allocation et alloue la mmoire pour en sauvegarder la structure (la mmoire est dsalloue dans CloseDB()).
FAT Read_FAT(DBase* db,int n);

//////////////////////////////////////////
//// FONCTIONS DE MANIPULATION DE CRC ////
//////////////////////////////////////////

//Calcule le CRC du segment de fichier dbutant  begin et de longueur size.
void Calc_CRC(DBase* db,long begin,long size,CRC32 &c);

//Calcule le CRC de la table de lien numro n.
void Calc_LT_CRC(DBase* db,int n,CRC32 &c);

//Calcule le CRC de la table d'allocation numro n.
void Calc_FAT_CRC(DBase* db,int n,CRC32 &c);

//Lit le CRC se trouvant  l'adresse offset.
void Read_CRC(DBase* db,long Offset,CRC32 &c);

//Renvoie le CRC inscrit au dbut de la table de lien numro n.
void Read_LT_CRC(DBase* db,int n,CRC32 &c);

//Renvoie le CRC inscrit au dbut de la table d'allocation numro n.
void Read_FAT_CRC(DBase* db,int n,CRC32 &c);

//Ecrit  l'adresse Offset le CRC spcifi.
void Write_CRC(DBase* db,long Offset,CRC32 &c);

//Ecrit le CRC de la table de lien numro n.
void Write_LT_CRC(DBase* db,int n,CRC32 &c);

//Ecrit le CRC de la table d'allocation numro n.
void Write_FAT_CRC(DBase* db,int n,CRC32 &c);

////////////////////////////////////////
//// FONCTIONS DE CALCUL D'ADRESSES ////
////////////////////////////////////////

//Renvoie l'adresse de l'en tte du fichier.
long GetOffset_Header();

//Renvoie l'adresse du dbut de la table de lien.
long GetOffset_First_LT();

//Renvoie l'adresse du CRC de la table de lien numro n.
long GetOffset_LT_CRC(DBase* db,int n);

//Renvoie l'adresse du dbut de la table de lien numro n.
long GetOffset_LT_Begin(DBase* db,int n);

//Renvoie l'adresse du dbut de la table d'allocation.
long GetOffset_First_FAT(DBase* db);

//Renvoie l'adresse du CRC de la table d'allocation numro n.
long GetOffset_FAT_CRC(DBase* db,int n);

//Renvoie l'adresse du nombre d'lment enregistr dans la table d'allocation numro n.
long GetOffset_FAT_NbItems(DBase* db,int n);

//Renvoie l'adresse du dbut de la table d'allocation numro n.
long GetOffset_FAT_Begin(DBase* db,int n);

//Renvoie l'adresse relative au dbut du fichier de la base de donnes elle-mme.
long GetOffset_DataBase(DBase* db);


